import{p as e,f as o}from"./flowDb-956e92f1-89489ae6.js";import{f as t,g as a}from"./styles-c10674c1-e4c5bdae.js";import{u as i}from"./index-c804b295.js";import"./graph-1ec4d266.js";import"./layout-1a08981e.js";import"./index-3862675e-0a0e97c9.js";import"./clone-cc738fa6.js";import"./edges-e0da2a9e-4bac0545.js";import"./createText-2e5e7dd3-57dc3038.js";import"./line-95f7f1d3.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-51d02e9e.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
